class Plugin():

    def __init__(self):
        self.name = self.__class__.__name__
        self.miner = None

    def before_hook(self, hook_name, payload):
        return True

    def set_miner(self, miner):
        self.miner = miner

    def notify(self, message, _type='info'):
        message = f"[{self.name}] {message}"
        self.miner.notify(message, _type)

    def __getattr__(self, key):
        if self.miner is None or key not in self.miner.__dict__:
            raise AttributeError(key)
        return getattr(self.miner, key)




import logging
import os
import time
from pathlib import Path

import functools
import torch
import tqdm
from IPython.core.display import HTML, display
import numpy as np
from . import drawers


class Miner(object):
    """The heart of minetorch

    Args:
        alchemistic_directory (string):
            The directory which minetorch will use to store everything in
        model (torch.nn.Module):
            Pytorch model optimizer (torch.optim.Optimizer): Pytorch optimizer
        loss_func (function):
            A special hook function to compute loss, the function receive 2 variable:
            * Miner: the miner object
            * Data: Batch data been yield by the loader
            return value of the hook function should be a float number of the loss
        code (str, optional):
            Defaults to "geass". It's a code name of one
            attempt. Assume one is doing kaggle competition and will try
            different models, parameters, optimizers... To keep results of every
            attempt, one should change the code name before tweaking things.
        train_dataloader (torch.utils.data.DataLoader):
            Pytorch dataloader
        val_dataloader (torch.utils.data.DataLoader, optional):
            Defaults to None, if no validation dataloader is provided, will skip validation
        resume (bool, optional):
            Defaults to True. Resume from last training, could be:
            * True: resume from the very last epochs
            * String: resume from the specified epochs
                          etc. `34`, `68` `best`
        eval_stride (int, optional):
            Defaults to 1. Validate every `eval_stride` epochs
        persist_stride (int, optional):
            Defaults to 1.
            Save model every `persist_stride` epochs
        drawer (minetorch.Drawer or string, optional):
            Defaults to matplotlib.
            If provide, Miner will draw training loss and validation loss
            curves, could be `tensorboard` or self implemented Drawer object
        hooks (dict, optional):
            Defaults to {}. Define hook functions.
        max_epochs ([type], optional):
            Defaults to None. How many epochs to train, None means unlimited.
        logging_format ([type], optional):
            Defaults to None. logging format
        trival ([Boolean], optional):
            Defaults to False. If true, both training and validation
            process will be breaked in 10 iterations
        plugins (list, optional):
            Defaults to []. This is actually a collection of `hooks`, do not set
            hooks and plugins the same time.
    """

    def __init__(self, alchemistic_directory, model, optimizer, loss_func, metrics=[],
                 code="geass", train_dataloader=None, val_dataloader=None,
                 resume=True, eval_stride=1, persist_stride=1, gpu=True,
                 drawer='matplotlib', hooks={}, max_epochs=None, statable={},
                 logging_format=None, trival=False, in_notebook=False, plugins=[],
                 logger=None):

        self.alchemistic_directory = alchemistic_directory
        self.code = code
        self.create_dirs()
        self.gpu = gpu
        self.devices = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.logger = logger
        if self.logger is None:
            self.set_logging_config(alchemistic_directory, code, logging_format)
            self.logger = logging
        self.create_drawer(drawer)
        self.models_dir = os.path.join(alchemistic_directory, code, 'models')
        self.in_notebook = in_notebook
        self.statable = statable

        self.model = model
        self.optimizer = optimizer
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader

        self.loss_func = loss_func
        self.metrics = metrics
        for metric in self.metrics:
            if not hasattr(metric, 'keywords'):
                metric = functools.partial(metric, func=lambda x: x, separate_class=True)

        self.resume = resume
        self.eval_stride = eval_stride
        self.persist_stride = persist_stride

        self.lowest_train_loss = float('inf')
        self.lowest_val_loss = float('inf')
        self.current_epoch = 0
        self.current_train_iteration = 0
        self.current_val_iteration = 0
        self.hook_funcs = hooks
        self.max_epochs = max_epochs
        self.metrics_metadata = {}

        self.init_model()
        self.trival = trival

        self._set_tqdm()
        self._check_statable()

        self.plugins = plugins
        for plugin in self.plugins:
            plugin.set_miner(self)

        self.status = 'init'
        self.call_hook_func('after_init')

    def _check_statable(self):
        for name, statable in self.statable.items():
            if not (hasattr(statable, 'state_dict') and hasattr(statable, 'load_state_dict')):
                raise Exception(f'The {name} is not a statable object')

    def _set_tqdm(self):
        if self.in_notebook:
            self.tqdm = tqdm.tqdm_notebook
        else:
            self.tqdm = lambda x: x

    def set_logging_config(self, alchemistic_directory, code, logging_format):
        self.log_dir = os.path.join(alchemistic_directory, code)
        log_file = os.path.join(self.log_dir, 'log.txt')
        logging_format = logging_format if logging_format is not None else \
            '%(levelname)s %(asctime)s %(message)s'
        logging.basicConfig(
            filename=log_file,
            format=logging_format,
            datefmt="%m-%d %H:%M:%S",
            level=logging.INFO
        )

    def create_drawer(self, drawer):
        if drawer == 'tensorboard':
            self.drawer = drawers.TensorboardDrawer(
                self.alchemistic_directory, self.code)
        elif drawer == 'matplotlib':
            self.drawer = drawers.MatplotlibDrawer(
                self.alchemistic_directory, self.code)
        else:
            self.drawer = drawer

    def notebook_output(self, message, _type='info'):
        type_config = {
            'info': ['💬', '#6f818a'],
            'success': ['✅', '#7cb305'],
            'error': ['❌', '#cf1322'],
            'warning': ['⚠️', '#d46b08'],
        }[_type]
        if self.in_notebook:
            display(HTML(
                f'<div style="font-size: 12px; color: {type_config[1]}">'
                f'⏰ {time.strftime("%b %d - %H:%M:%S")} >>> '
                f'{type_config[0]} {message}'
                '</div>'
            ))

    def notebook_divide(self, message):
        if self.in_notebook:
            display(HTML(
                '<div style="display: flex; justify-content: center;">'
                f'<h3 style="color: #7cb305; border-bottom: 4px dashed #91d5ff; padding-bottom: 6px;">{message}</h3>'
                '</div>'
            ))

    def init_model(self):
        """resume from some checkpoint
        """
        if isinstance(self.model, torch.nn.DataParallel):
            raise Exception(
                'Don\'t parallel the model yourself, instead, if the '
                '`gpu` option is true(default), Minetorch will do this for you.'
            )

        if self.resume is True:
            # resume from the newest model
            if self.model_file_path('latest') is not None:
                checkpoint_path = self.model_file_path('latest')
            else:
                checkpoint_path = None
                msg = ('Could not find checkpoint to resume, '
                       'train from scratch')
                self.logger.warning(msg)
                self.notebook_output(msg, _type='warning')
        elif isinstance(self.resume, str):
            checkpoint_path = self.model_file_path(self.resume)
        elif isinstance(self.resume, int):
            checkpoint_path = self.model_file_path(str(self.resume))
        else:
            checkpoint_path = None

        if self.resume is not True and self.resume and checkpoint_path is None:
            # user has specified a none existed model, should raise a error
            raise Exception(f"Could not find model {self.resume}")

        if checkpoint_path is not None:
            msg = f"Start to load checkpoint {checkpoint_path}"
            self.logger.info(msg)
            self.notebook_output(msg)
            checkpoint = torch.load(checkpoint_path)
            self.current_epoch = checkpoint['epoch']
            self.current_train_iteration = checkpoint['train_iteration']
            self.current_val_iteration = checkpoint['val_iteration']
            self.lowest_train_loss = checkpoint['lowest_train_loss']
            self.lowest_val_loss = checkpoint['lowest_val_loss']

            try:
                self.model.load_state_dict(checkpoint['state_dict'], strict=True)
            except:
                msg = ('load checkpoint failed, the state in the '
                       'checkpoint is not matched with the model, '
                       'try to reload checkpoint with unstrict mode')
                self.logger.warning(msg)
                self.notebook_output(msg)
                self.model.load_state_dict(checkpoint['state_dict'], strict=False)

            if 'optimizer' in checkpoint:
                try:
                    self.optimizer.load_state_dict(checkpoint['optimizer'])
                except:
                    msg = ('load optimizer state failed, will skip this error and continue, '
                            'stop the process if it is not expected')
                    self.logger.warning(msg)
                    self.notebook_output(msg)

            if (self.drawer is not None) and ('drawer_state' in checkpoint):
                self.drawer.set_state(checkpoint['drawer_state'])

            if 'statable' in checkpoint:
                for name, statable in self.statable.items():
                    if name not in checkpoint['statable']:
                        continue
                    statable.load_state_dict(checkpoint['statable'][name])
            msg = 'checkpoint loaded'
            self.notebook_output(msg, _type='success')

        if self.gpu:
            gpu_count = torch.cuda.device_count()
            if gpu_count == 0:
                self.notify('no GPU detected, will train on CPU.')
            else:
                self.notify(f'found {gpu_count} GPUs, will use all of them to train')
                devices = list(map(lambda x: f'cuda:{x}', range(gpu_count)))
                self.model.cuda()
                self.model = torch.nn.DataParallel(self.model, devices)

    def notify(self, message, _type='info'):
        getattr(self.logger, _type)(message)
        self.notebook_output(message, _type)

    def call_hook_func(self, name, **payload):
        if name in self.hook_funcs:
            self.hook_funcs[name](payload)
        else:
            for plugin in self.plugins:
                if not plugin.before_hook(name, payload):
                    continue
                if hasattr(plugin, name):
                    getattr(plugin, name)(payload)

    def train(self):
        """start to train the model
        """
        while True:
            self.current_epoch += 1
            self.call_hook_func('before_epoch_start', epoch=self.current_epoch)
            self.notebook_divide(f'Epoch {self.current_epoch}')
            self.model.train()
            train_iters = len(self.train_dataloader)

            total_train_loss = 0
            total_train_metrics = {}
            self.notebook_output(f'start to train epoch {self.current_epoch}')
            for index, data in enumerate(self.tqdm(self.train_dataloader)):
                if self.trival is True and index == 10:
                    break
                train_loss, train_metrics = self.run_train_iteration(index, data, train_iters)
                total_train_loss += train_loss
                for metric in train_metrics:
                    if metric not in total_train_metrics:
                        total_train_metrics[metric] = 0
                    total_train_metrics[metric] += train_metrics[metric]
            for i, j in enumerate(total_train_metrics):
                total_train_metrics[j] = total_train_metrics[j] / train_iters
                total_train_metrics[j] = self.metrics[i].keywords['func'](total_train_metrics[j])

            total_train_loss = total_train_loss / train_iters
            self.notebook_output(f'training of epoch {self.current_epoch} finished, '
                                 f'loss is {total_train_loss}')

            total_val_loss = 0
            total_val_metrics = {}
            if self.val_dataloader is not None:
                val_iters = len(self.val_dataloader)
                with torch.set_grad_enabled(False):
                    self.model.eval()
                    self.notebook_output(f'validate epoch {self.current_epoch}')
                    for index, data in enumerate(self.tqdm(self.val_dataloader)):
                        if self.trival is True and index == 10:
                            break
                        val_loss, val_metrics = self.run_val_iteration(index, data, val_iters)
                        total_val_loss += val_loss
                        for metric in val_metrics:
                            if metric not in total_val_metrics.keys():
                                total_val_metrics[metric] = 0
                                total_val_metrics[metric] += val_metrics[metric]
                            else:
                                total_val_metrics[metric] += val_metrics[metric]
                for i, j in enumerate(total_val_metrics):
                    total_val_metrics[j] = total_val_metrics[j] / val_iters
                    total_val_metrics[j] = self.metrics[i].keywords['func'](total_val_metrics[j])

                total_val_loss = total_val_loss / val_iters
                self.notebook_output(f'validation of epoch {self.current_epoch}'
                                     f'finished, loss is {total_val_loss}')
            self.call_hook_func(
                'after_epoch_end',
                train_loss=total_train_loss,
                val_loss=total_val_loss,
                train_metric=train_metrics,
                val_metric=val_metrics,
                epoch=self.current_epoch
                )

            if self.drawer is not None:
                self.drawer.scalars(
                    {'train': total_train_loss, 'val': total_val_loss}, 'loss'
                )
                for metric in self.metrics:
                    if total_train_metrics[metric.keywords['func'].__name__].shape.numel() != 1:
                        for i in range(total_train_metrics[metric.keywords['func'].__name__].shape.numel()):
                            self.drawer.scalars(
                                {
                                    'train': total_train_metrics[metric.keywords['func'].__name__][i],
                                    'val': total_val_metrics[metric.keywords['func'].__name__][i]
                                    },
                                metric.keywords['func'].__name__+'_class_{}'.format(i+1)
                                )
                    else:
                        self.drawer.scalars(
                            {'train': total_train_metrics[metric.keywords['func'].__name__],
                            'val': total_val_metrics[metric.keywords['func'].__name__]},
                            metric.keywords['func'].__name__
                            )

            if total_train_loss < self.lowest_train_loss:
                self.lowest_train_loss = total_train_loss

            if total_val_loss < self.lowest_val_loss:
                message = ('current val loss {} is lower than lowest {}, '
                           'persist this model as best one'.format(
                            total_val_loss, self.lowest_val_loss))
                self.notebook_output(f'{message}', _type='success')
                self.logger.info(message)

                self.lowest_val_loss = total_val_loss
                self.persist('best')
            self.persist('latest')

            if not self.current_epoch % self.persist_stride:
                self.persist('epoch_{}'.format(self.current_epoch))

            if self.max_epochs is not None and self.current_epoch >= self.max_epochs:
                self.call_hook_func('before_quit')
                self.logger.info('exceed max epochs, quit!')
                break

    def run_train_iteration(self, index, data, train_iters):
        self.status = 'train'
        self.current_train_iteration += 1
        self.call_hook_func(
            'before_train_iteration_start',
            data=data,
            index=index,
            total_iters=train_iters,
            iteration=self.current_train_iteration
            )
        batch_size = data[0].shape[0]
        predict = self.model(data[0].to(self.devices))
        loss = self.loss_func(predict, data[1].to(self.devices))
        train_metrics = {}
        for metric in self.metrics: # iterate metrics specified by users
            train_metrics[metric.keywords['func'].__name__] = 0
            for batch in range(batch_size): # iterate data from the dataloader
                values, _ = metric(predict[batch].detach().cpu(), data[1][batch]) # return confusion_matrix and specific functions
                train_metrics[metric.keywords['func'].__name__] += values  # predict.shape = [B,C,H,W]
            train_metrics[metric.keywords['func'].__name__] /= batch_size

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        loss = loss.detach().cpu().item()
        self.logger.info('[train {}/{}/{}] loss {}'.format(
            self.current_epoch, index, train_iters, loss))
        if loss < self.lowest_train_loss:
            self.lowest_train_loss = loss

        self.call_hook_func(
            'after_train_iteration_end',
            loss=loss,
            data=data,
            index=index,
            total_iters=train_iters,
            iteration=self.current_train_iteration
            )
        return loss, train_metrics

    def run_val_iteration(self, index, data, val_iters):
        self.status = 'val'
        self.current_val_iteration += 1
        self.call_hook_func(
            'before_val_iteration_start',
            data=data,
            index=index,
            total_iters=val_iters,
            iteration=self.current_val_iteration
            )
        batch_size = data[0].shape[0]
        predict = self.model(data[0].to(self.devices))
        loss = self.loss_func(predict, data[1].to(self.devices))
        val_metrics = {}
        for metric in self.metrics:
            val_metrics[metric.keywords['func'].__name__] = 0
            for batch in range(batch_size):
                values, func = metric(predict[batch].detach().cpu(), data[1][batch])
                val_metrics[metric.keywords['func'].__name__] += values  # predict.shape = [B,C,H,W]
            val_metrics[metric.keywords['func'].__name__] /= batch_size
        loss = loss.detach().cpu().item()
        self.logger.info('[val {}/{}/{}] loss {}'.format(
            self.current_epoch, index, val_iters, loss))

        self.call_hook_func(
            'after_val_iteration_ended',
            predicts=predict,
            loss=loss,
            data=data,
            index=index,
            total_iters=val_iters,
            iteration=self.current_val_iteration
            )
        return loss, val_metrics

    def persist(self, name):
        """save the model to disk
        """
        self.call_hook_func('before_checkpoint_persisted')
        if self.drawer is not None:
            drawer_state = self.drawer.get_state()
        else:
            drawer_state = {}

        if isinstance(self.model, torch.nn.DataParallel):
            model_state_dict = self.model.module.state_dict()
        else:
            model_state_dict = self.model.state_dict()

        state = {
            'state_dict': model_state_dict,
            'optimizer': self.optimizer.state_dict(),
            'epoch': self.current_epoch,
            'train_iteration': self.current_train_iteration,
            'val_iteration': self.current_val_iteration,
            'lowest_train_loss': self.lowest_train_loss,
            'lowest_val_loss': self.lowest_val_loss,
            'drawer_state': drawer_state,
            'statable': {}
        }

        for statable_name, statable in self.statable.items():
            state['statable'][statable_name] = statable.state_dict()

        modelpath = self.standard_model_path(name)
        torch.save(state, modelpath)
        message = f'save checkpoint to {self.standard_model_path(name)}'
        self.logger.info(message)
        self.notebook_output(message)
        self.call_hook_func('after_checkpoint_persisted', modelpath=modelpath)

    def standard_model_path(self, model_name):
        return os.path.join(self.models_dir, f'{model_name}.pth.tar')

    def model_file_path(self, model_name):
        model_name_path = Path(model_name)
        models_dir_path = Path(self.models_dir)

        search_paths = [
            model_name_path,
            models_dir_path / model_name_path,
            models_dir_path / f'{model_name}.pth.tar',
            models_dir_path / f'epoch_{model_name}.pth.tar',
        ]

        for path in search_paths:
            if path.is_file():
                return path.resolve()

        return None

    # TODO: implement methods below
    def graceful_stop(self):
        """stop train and exist after this epoch
        """
        pass

    def save_and_stop(self):
        """save the model immediately and stop training
        """
        pass

    def create_dirs(self):
        """Create directories
        """
        self.create_dir('')
        self.create_dir(self.code)
        self.create_dir(self.code, 'models')

    def create_dir(self, *args):
        """Create directory
        """
        current_dir = self.alchemistic_directory
        for dir_name in args:
            current_dir = os.path.join(current_dir, dir_name)
            if not os.path.isdir(current_dir):
                os.mkdir(current_dir)



from pathlib import Path

import cv2
from torchvision import transforms
import inspect


class ProcessorBundler:

    def __init__(self, processors=[]):
        self.processors = processors

    def add_processors(self, processors=[]):
        self.processors += processors

    def prepend_processors(self, processors=[]):
        self.processors = processors + self.processors

    def __call__(self, data):
        for (processor, columns) in self.processors:
            if not isinstance(columns, tuple):
                columns = (columns,)

            for column in columns:
                data[column] = processor(data[column])
        return data


class Processor:

    def __call__(self):
        raise NotImplementedError()

    def source_file(self):
        return inspect.getsourcefile(self.__class__)

    def source_lineno(self):
        return inspect.getsourcelines(self.__class__)[1]


class PathMaker(Processor):
    def __init__(self, base_directory, extname=None):
        self.base_directory = Path(base_directory)
        self.extname = extname

    def __call__(self, filename):
        path = self.base_directory / filename
        if self.extname is not None:
            path = path.with_suffix(self.extname)
        return path.as_posix()


class ImageLoader(Processor):

    def __init__(self, image_size=None):
        self.image_size = image_size

    def __call__(self, filename):
        image = cv2.imread(filename)
        if self.image_size is not None:
            image = cv2.resize(image, self.image_size)
        return image


class Augmentor(Processor):
    """This is a tiny wrapper of albumentations
    see: https://github.com/albu/albumentations
    """

    def __init__(self, aug):
        """
        Args:
            aug (albumentations.core.composition.Compose)
        """
        self.aug = aug

    def __call__(self, tensor):
        return self.aug(image=tensor)['image']


class ImagenetNormalizor(Processor):

    def __init__(self):
        self.normalize = transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225]
        )

    def __call__(self, tensor):
        return self.normalize(tensor)



import pandas as pd
from torch.utils.data import Dataset


class MineDataset(Dataset):

    def __init__(self, data, processor=None):
        self.data = data
        self.processor = processor

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        # processor will modifed data inplace, `copy()`
        # to prevent self.data being modified
        data = self.data[index].copy()
        if self.processor is not None:
            data = self.processor(data)
        return data


class CsvDataset(MineDataset):

    def __init__(self, csv_file_path, processor=None):
        data = self.generate_data(csv_file_path)
        super().__init__(data, processor)

    def generate_data(self, csv_file_path):
        df = pd.read_csv(csv_file_path)
        return df.values.tolist()



import torch
import torch.nn.functional as F
from torch.autograd import Variable


# Below loss functions are referenced at https://www.kaggle.com/bigironsphere/loss-function-library-keras-pytorch
def dice_loss(inputs, targets, smooth=1):
    inputs = torch.sigmoid(inputs)
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    intersection = (inputs * targets).sum()
    dice = (2.*intersection + smooth)/(inputs.sum() + targets.sum() + smooth)
    return 1 - dice


def dicebce_loss(inputs, targets, smooth=1):
    inputs = torch.sigmoid(inputs)
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    intersection = (inputs * targets).sum()
    dice_loss = 1 - (2.*intersection + smooth)/(inputs.sum() + targets.sum() + smooth)
    bce = F.binary_cross_entropy(inputs, targets, reduction='mean')
    dice_bce = bce + dice_loss
    return dice_bce


def iou_loss(inputs, targets, smooth=1):
    inputs = torch.sigmoid(inputs)
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    intersection = (inputs * targets).sum()
    total = (inputs + targets).sum()
    union = total - intersection
    iou = (intersection + smooth)/(union + smooth)
    return 1 - iou


def focal_loss(inputs, targets, alpha=0.8, gamma=2, smooth=1):
    inputs = torch.sigmoid(inputs)
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    bce = F.binary_cross_entropy(inputs, targets, reduction='mean')
    bce_exp = torch.exp(-bce)
    focal_loss = alpha * (1-bce_exp)**gamma * bce
    return focal_loss


def tversky_loss(inputs, targets, smooth=1, alpha=0.5, beta=0.5):
    inputs = torch.sigmoid(inputs)
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    tp = (inputs * targets).sum()
    fp = ((1-targets) * inputs).sum()
    fn = (targets * (1-inputs)).sum()
    tversky = (tp + smooth) / (tp + alpha * fp + beta * fn + smooth)
    return 1 - tversky


def focaltversky_loss(inputs, targets, smooth=1, alpha=0.5, beta=0.5, gamma=1):
    inputs = torch.sigmoid(inputs)
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    tp = (inputs * targets).sum()
    fp = ((1-targets) * inputs).sum()
    fn = (targets * (1-inputs)).sum()
    tversky = (tp + smooth) / (tp + alpha * fp + beta * fn + smooth)
    focaltversky = (1 - tversky)**gamma
    return focaltversky


def lovaszhingeLoss(inputs, targets):
    inputs = torch.sigmoid(inputs)
    lovasz = lovasz_hinge(inputs, targets, per_image=False)
    return lovasz


def flatten_binary_scores(scores, labels, ignore=None):
    """
    Flattens predictions in the batch (binary case)
    Remove labels equal to 'ignore'
    """
    scores = scores.view(-1)
    labels = labels.view(-1)
    if ignore is None:
        return scores, labels
    valid = (labels != ignore)
    vscores = scores[valid]
    vlabels = labels[valid]
    return vscores, vlabels


def lovasz_grad(gt_sorted):
    """
    Computes gradient of the Lovasz extension w.r.t sorted errors
    See Alg. 1 in paper
    """
    p = len(gt_sorted)
    gts = gt_sorted.sum()
    intersection = gts - gt_sorted.float().cumsum(0)
    union = gts + (1 - gt_sorted).float().cumsum(0)
    jaccard = 1. - intersection / union
    if p > 1:
        jaccard[1:p] = jaccard[1:p] - jaccard[0:-1]
    return jaccard


def lovasz_hinge(logits, labels, per_image=True, ignore=None):
    """
    Binary Lovasz hinge loss
      logits: [B, H, W] Variable, logits at each pixel (between - infty and + infty)
      labels: [B, H, W] Tensor, binary ground truth masks (0 or 1)
      per_image: compute the loss per image instead of per batch
      ignore: void class id
    """
    if per_image:
        loss = mean(
            lovasz_hinge_flat(
                *flatten_binary_scores(
                    log.unsqueeze(0),
                    lab.unsqueeze(0),
                    ignore)
                    ) for log, lab in zip(logits, labels))
    else:
        loss = lovasz_hinge_flat(
            *flatten_binary_scores(
                logits,
                labels,
                ignore)
                )
    return loss


def lovasz_hinge_flat(logits, labels):
    """
    Binary Lovasz hinge loss
      logits: [P] Variable, logits at each prediction (between - infty and + infty)
      labels: [P] Tensor, binary ground truth labels (0 or 1)
      ignore: label to ignore
    """
    if len(labels) == 0:
        # only void pixels, the gradients should be 0
        return logits.sum() * 0.
    signs = 2. * labels.float() - 1.
    errors = (1. - logits * Variable(signs))
    errors_sorted, perm = torch.sort(errors, dim=0, descending=True)
    perm = perm.data
    gt_sorted = labels[perm]
    grad = lovasz_grad(gt_sorted)
    loss = torch.dot(F.relu(errors_sorted), Variable(grad))
    return loss


def lovasz_softmax(probas, labels, classes='present', per_image=False, ignore=None):
    """
    Multi-class Lovasz-Softmax loss
      probas: [B, C, H, W] Variable, class probabilities at each prediction (between 0 and 1).
              Interpreted as binary (sigmoid) output with outputs of size [B, H, W].
      labels: [B, H, W] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
      per_image: compute the loss per image instead of per batch
      ignore: void class labels
    """
    if per_image:
        loss = mean(
            lovasz_softmax_flat(
                *flatten_probas(
                    prob.unsqueeze(0),
                    lab.unsqueeze(0),
                    ignore
                    ),
                classes=classes) for prob, lab in zip(probas, labels))
    else:
        loss = lovasz_softmax_flat(*flatten_probas(probas, labels, ignore), classes=classes)
    return loss


def lovasz_softmax_flat(probas, labels, classes='present'):
    """
    Multi-class Lovasz-Softmax loss
      probas: [P, C] Variable, class probabilities at each prediction (between 0 and 1)
      labels: [P] Tensor, ground truth labels (between 0 and C - 1)
      classes: 'all' for all, 'present' for classes present in labels, or a list of classes to average.
    """
    if probas.numel() == 0:
        # only void pixels, the gradients should be 0
        return probas * 0.
    C = probas.size(1)
    losses = []
    class_to_sum = list(range(C)) if classes in ['all', 'present'] else classes
    for c in class_to_sum:
        fg = (labels == c).float()
        if (classes == 'present' and fg.sum() == 0):
            continue
        if C == 1:
            if len(classes) > 1:
                raise ValueError('Sigmoid output possible only with 1 class')
            class_pred = probas[:, 0]
        else:
            class_pred = probas[:, c]
        errors = (Variable(fg) - class_pred).abs()
        errors_sorted, perm = torch.sort(errors, 0, descending=True)
        perm = perm.data
        fg_sorted = fg[perm]
        losses.append(torch.dot(errors_sorted, Variable(lovasz_grad(fg_sorted))))
    return mean(losses)


def combo_loss(inputs, targets, smooth=1, alpha=0.5, ce_ratio=0.5):
    e = 1e-7
    inputs = inputs.view(-1)
    targets = targets.view(-1)
    intersection = (inputs * targets).sum()
    dice = (2. * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)
    inputs = torch.clamp(inputs, e, 1.0 - e)
    out = - (alpha * ((targets * torch.log(inputs)) + ((1 - alpha) * (1.0 - targets) * torch.log(1.0 - inputs))))
    weighted_ce = out.mean(-1)
    combo = (ce_ratio * weighted_ce) - ((1 - ce_ratio) * dice)
    return combo



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='Minetorch',
    description='A tools collection for pytorch users',
    version='0.4.13',
    packages=['minetorch'],
    url="https://github.com/louis-she/minetorch",
    author='louis',
    author_email='chenglu.she@gmail.com',
    keywords='pytorch minecraft',
    install_requires=['tensorboardX', 'torch', 'matplotlib', 'albumentations'],
    entry_points = {
        'console_scripts': ['minetorch=minetorch.command_line:main'],
    }
)





import torch
import functools


def shape_norm(logits, targets):
    # Expect one of the following shapes [classes, height, width], [height, width], [classes].
    assert logits.shape == targets.shape, "Inputs not match with targets on shape"
    assert len(logits.shape) <= 3, 'Expect one of the following shapes [classes, height, width], [height, width], [classes].'
    if len(logits.shape) == 1:
        logits = logits.unsqueeze(1).unsqueeze(1)
        targets = targets.unsqueeze(1).unsqueeze(1)
    elif len(logits.shape) == 2:
        logits = logits.unsqueeze(0)
        targets = targets.unsqueeze(0)
    logits = torch.sigmoid(logits)
    return logits, targets


def compute_mse(logits, targets, separate_class=False):
    logits, targets = shape_norm(logits, targets)
    error = (abs(logits - targets) ** 2).double().sum((1, 2))
    total = logits.view(logits.shape[0], -1).shape[1]
    mse = error / total
    if separate_class:
        mse = mse
    else:
        mse = mse.mean()
    return mse


def compute_mae(logits, targets, separate_class=False):
    logits, targets = shape_norm(logits, targets)
    error = (abs(logits - targets)).double().sum((1, 2))
    total = logits.view(logits.shape[0], -1).shape[1]
    mae = error / total
    if separate_class:
        mae = mae
    else:
        mae = mae.mean()
    return mae


def confusion_matrix(logits, targets, threshold=0.5, separate_class=False, func=lambda x: x):
    logits, targets = shape_norm(logits, targets)
    logits = logits > threshold
    targets = targets > 0.5
    tp = ((logits == 1) & (targets == 1)).double().sum((1, 2)).unsqueeze(0)
    fp = ((logits == 1) & (targets == 0)).double().sum((1, 2)).unsqueeze(0)
    fn = ((logits == 0) & (targets == 1)).double().sum((1, 2)).unsqueeze(0)
    tn = ((logits == 0) & (targets == 0)).double().sum((1, 2)).unsqueeze(0)
    return torch.cat((tp, fp, fn, tn), 0), func


def compute_precision(c_matrix):
    return c_matrix[0] / (c_matrix[0] + c_matrix[1])


def compute_recall(c_matrix):
    return c_matrix[0] / (c_matrix[0] + c_matrix[2])


def compute_accuracy(c_matrix):
    return (c_matrix[0] + c_matrix[3]) / (c_matrix[0] + c_matrix[1] + c_matrix[2] + c_matrix[3])


def compute_dice(c_matrix):
    smooth = 1e-7
    return (2 * c_matrix[0] + smooth) / (2 * c_matrix[0] + c_matrix[1] + c_matrix[2] + smooth)


def compute_iou(c_matrix):
    smooth = 1e-7
    return (c_matrix[0] + smooth) / (c_matrix[0] + c_matrix[1] + c_matrix[2] + smooth)


precision = functools.partial(confusion_matrix, func=compute_precision, separate_class=True)
recall = functools.partial(confusion_matrix, func=compute_recall, separate_class=True)
accuracy = functools.partial(confusion_matrix, func=compute_accuracy, separate_class=True)
dice = functools.partial(confusion_matrix, func=compute_dice, separate_class=True)
iou = functools.partial(confusion_matrix, func=compute_iou, separate_class=True)



import os
from matplotlib.figure import Figure
from tensorboardX import SummaryWriter
import _pickle as pickle
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas


class Drawer():
    """To vistualize everything in training process
    """

    def __init__(self, alchemistic_directory, code, state=None):
        """Constructor

        Args:
            alchemistic_directory (string):
                same as trainer's alchemistic_directory
            code (string):
                same as trainer's alchemistic_directory
            graph (state, optional):
                Defaults to None. Since we could draw multiple graph during
                training, the state records the current position of each graph
                The keys are the name of the graphs and values are current
                positions.
        """
        self.step_file = os.path.join(alchemistic_directory, code, '.drawer_step')
        self.code = code
        self.alchemistic_directory = alchemistic_directory

        if state is None:
            self.state = {}
        else:
            self.state = state

    def scalars(self, value, graph):
        """Plot different scalars on a graph

        Args:
            value (dict):
                scalar to plot
            graph (string):
                graph name
        """
        raise NotImplementedError()

    def scalar(self, value, graph):
        """Plot one scalar on a graph

        Args:
            value (float):
                scalar to plot
            graph (string):
                graph name
        """
        self.scalars({graph: value}, graph)

    def get_state(self):
        """Return current state(counter) of the Drawer
        """
        return self.state

    def set_state(self, state):
        """Set current state(counter) to state
        """
        self.state = state


class TensorboardDrawer(Drawer):
    """To vistualize everything in training process using tensorboard
    """

    def __init__(self, alchemistic_directory, code, step=None):
        super().__init__(alchemistic_directory, code, step)
        self.writer = SummaryWriter(log_dir=os.path.join(
            alchemistic_directory, code
        ))

    def scalars(self, value, graph):
        """Add a scalar on a graph

        Args:
            value (dict):
                scalars to put on the graph
            graph (string):
                graph name
        """
        if graph not in self.state:
            self.state[graph] = 0
        key = '{}/{}'.format(self.code, graph)
        if isinstance(value, dict):
            self.writer.add_scalars(key, value, self.state[graph])
        else:
            self.writer.add_scalar(key, value, self.state[graph])
        self.state[graph] += 1


class MatplotlibDrawer(Drawer):

    def __init__(self, alchemistic_directory, code, state=None):
        super().__init__(alchemistic_directory, code, state)
        self.graph_dir = os.path.join(alchemistic_directory, code, 'graphs')
        self.data_file = os.path.join(self.graph_dir, '.graphs.pickle')
        self.colors = ['blue', 'orange', 'green', 'red', 'purple',
                       'brown', 'pink', 'gray', 'olive', 'cyan']
        if not os.path.isdir(self.graph_dir):
            os.mkdir(self.graph_dir)
        if os.path.isfile(self.data_file):
            with open(self.data_file, 'rb') as f:
                self.graph_data = pickle.load(f)

    def _update_state(self, values, graph):
        if graph not in self.state or not isinstance(self.state[graph], dict):
            self.state[graph] = {}
        for key in values:
            if key not in self.state[graph]:
                self.state[graph][key] = []
            self.state[graph][key].append(values[key])
        with open(self.data_file, 'wb') as f:
            pickle.dump(self.state, f)

    def _save_png(self, graph):
        png_file = os.path.join(self.graph_dir, graph + ".png")
        fig = Figure()
        FigureCanvas(fig)
        ax = fig.add_subplot(1, 1, 1)
        ax.grid(True)
        for index, curve in enumerate(self.state[graph]):
            ax.plot(self.state[graph][curve], label=curve, color=self.colors[index])

        ax.legend(loc='upper left')
        fig.savefig(png_file)

    def scalars(self, values, graph):
        """Add a scalar on a graph

        Args:
            value (dict):
                scalars to put on the graph
            graph (string):
                graph name
        """
        self._update_state(values, graph)
        self._save_png(graph)
